#!/usr/bin/perl -I../lib
#
# $Id$
#

# Copyright INRA

# Celine.Noirot@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr
# Christine.Gaspin@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based system for detecting, clustering and annotating non protein
# coding RNA genes.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.



=head1 NAME

 LeARN.pl - From a multifasta file, build a learn rna database.

=head1 SYNOPSIS

 $LEARN_DIR/bin/RunProgram.pl --cfg $LEARN_DIR/cfg/Learn.cfg --input_fasta $LEARN_DIR/data/multifasta 

=head1 DESCRIPTION

	--help                          this message
	--cfg  filename                 full path of the configuration file
	--path_old_release dirname      path of old release
	--path_new_release dirnamr      path of new release (to create)
	--input_fasta filename          multifasta input file of one species
	--overlaps filename             file contain overlap data : [seq1 start1 end1 strand1 seq2 start2 end2 strand2]
	--species quoted_string         name of the organism , eg: 'Medicago truncatula'
	--upgraded_sequences filename 	is some version of BAC are upgraded specify the file wich contain list of  [old_bac new_bac]
	--log filename                  log file


=cut


BEGIN
{
	my ($dir,$file) = $0 =~ /(.+)\/(.+)/;
	unshift (@INC,"$dir/../lib");
}

use strict;
use warnings;
use Learn;
use ParamParser;
use Logger;
our $LOGGER_LEVEL_EXEC = 'DE';
#our $o_param ;
MAIN:
{
	# Tout les parametres sont dans le cfg
	my $o_param = New ParamParser('GETOPTLONG',"cfg=s", "path_old_release=s","path_new_release=s","input_fasta=s", "overlaps=s",  "species=s", "upgraded_sequences=s", "log=s" );
        
	$o_param->SetUsage(my $usage=sub { &Usage(); } );
	#tester RFAM env 

	$o_param->Update('ENV','A');
	$o_param->AssertDefined('LEARN_DIR');
	$o_param->SetUnlessDefined('cfg',$o_param->Get('LEARN_DIR')."/cfg/Learn.cfg");
        $o_param->AssertFileExists('cfg');
	$o_param->Update($o_param->Get('cfg'),'A');
	$o_param->AssertFileExists('input_fasta');
	my $fasta=$o_param->Get('input_fasta');

	$o_param->SetUnlessDefined('RFAM_DIR',$o_param->Get('path_old_release')."/".$o_param->Get('path_db_rfam') );
	$o_param->AssertDirExists('RFAM_DIR');
	$ENV{'RFAM_DIR'}=$o_param->Get('RFAM_DIR');
        
	my $new_release=$o_param->Get('path_new_release');
	if ( -d $new_release )
	{
		die ("New release already exist : $new_release \n");
	}
	
	$o_param->SetUnlessDefined('LOGGER_LEVEL_EXEC',$LOGGER_LEVEL_EXEC);
	my $logger = New Logger ($o_param->Get('log'),$LOGGER_LEVEL_EXEC,'NOCOMPRESSION');
	$logger->Trace("Begin Learn :\n",$logger->CalcStamp,"DE");
# 
	my $application= New Learn($o_param, \$logger);
	$o_param->SetUnlessDefined('createtree',1);
    	$o_param->SetUnlessDefined('db_opt',"ln");
	
	$application->Initialize($o_param->Get('createtree'), $o_param->Get('db_opt'),$o_param->Get('overlaps'), $o_param->Get('upgraded_sequences')); # crée l'arborescence

	$application->CreatePrograms(); # blastn / rfam_scan / trnaScan-SE / rnafold / mfold
        # set use repository to all scan programs.
	$application->SetScanProgramRepository($o_param->Get('repository_dir'), $o_param->Get('overwrite_repository'));
        # set use repository to program of folding for rna (default mfold).
	${$application->GetProgram("R_FOLD")}->SetRepository($o_param->Get('repository_dir'), $o_param->Get('overwrite_repository'));

	$application->CopyRflDb();
	$application->LoadRelease($o_param->Get('old_release'));
        
	$application->LoadSequences($fasta);
	$application->AnalyseSequences("LoadRna"); 
	$application->ComputeAllRnaStructure();

	${$application->GetProgram("ISCAN")}->UnsetLoadFromRepository();
	$application->BuildFamily();
	$application->DetectFamilyToMerge('rfam');
	$application->ImproveSSFamilies();
	$application->DumpAnnotationBySequence();
	#$application->PrintFamilies();
	#$application->PrintRnas();
	$application->SaveRnasToRnaml();
	$application->SaveFamiliesToRnaml();

	$application->DumpStatus($o_param->Get('path_new_release')."/".$o_param->Get('status_file'), $o_param->Get('path_old_release')."/".$o_param->Get('status_file'));
	if ($o_param->Get('clean'))
	{
		foreach my $tmprep ('path_db_learn_family_tmp','path_db_learn_rna_tmp' )
		{
			my $dir = $o_param->Get('path_new_release')."/".$o_param->Get($tmprep);
			if ( -d $dir )
			{
				my $cmd = "find  $dir/  -type d  -a ! -name tmp -exec rm -rf '{}' \\; 2> /dev/null";
				#print $cmd;
				system $cmd;
			}
		}
	}
	$logger->Trace("End Learn :\n",$logger->CalcStamp,"DE");
}

=item Usage

	$o_param->Usage();
	Print the usage of the program

=cut

sub Usage
{
print STDERR<<END
Usage:$0	

[Mandatory]
	--input_fasta filename     filename        multifasta input file of one species
[Optionnal]
	--help                                     this message
	--cfg                      filename        [LEARN_DIR/cfg/Learn.cfg] full path of the configuration file
	--path_old_release dirname directory       path of old release [default set in cfg]
	--path_new_release dirname directory       path of new release (to be create) [default set in cfg]
	--overlaps filename        filename        file contain overlap data 
                                                   [seq1 start1 end1 strand1 seq2 start2 end2 strand2]
	--species                  quoted_string   name of the organism , eg: 'Medicago truncatula'
	--upgraded_sequences       filename 	   If some version of BAC are upgraded specify the file wich contain list of  
                                                   [old_bac new_bac]
	--log filename             filename        [LEARN_DIR/log/learn.log] log file
END
}
