#
# $Id$
#

# Copyright INRA

# Celine.Noirot@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr
# Christine.Gaspin@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based system for detecting, clustering and annotating non protein
# coding RNA genes.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
 
#
# 	Celine.Noirot@toulouse.inra.fr
#	Last updated: October 19, 2005
#

package Program;

=pod

=head1 NAME

Program - It's an object which execute program

=head1 SYNOPSIS

	my $prog = New Program($program_name,$program_path,$program_param);	
	
example: 
	use Program;
	my $prog = New Program('blast','/usr/local/blastn',"-d database");

	#set input and output files
	$prog->SetInOut('/tmp/my_seq.fa','/tmp/my_seq.blast');

	#run the command 
	$prog->Run;

The package recognize the program_name :
	blast, rnafold, rnaeval, rnaalifold, rnaplot, rnaploteval, multalin, mfold, cmbuild, cmsearch, formatdb, clustal, AnchoRna

	
=head1 DESCRIPTION


=cut

use strict;
use warnings;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use Learn;

our $R_LOGGER;
our $LOGGER_STAMP;

BEGIN
{
    our $VERSION = "0.1";
}
use Var;


=head2 function New

 Title        : New
 Usage        : New Program($program_name,$program_path,$program_param);
 Prerequisite : none
 Function     : Create an object Program
 Returns      : An object
 Args         : $program_name : name of the program eg : blast, rnafold, multalin
 								name is use to Build the command line for special program
 				$program_path : path of the program  eg : /usr/local/blastn
				$program_param : parameters of the program
 Globals      : none

=cut 

sub New
{
    my ($class, $name, $prg, $param, $type, $version, $r_log, $log_stamp) = @_;
	$R_LOGGER=$r_log;
	$LOGGER_STAMP=$log_stamp;
	my $self = {
                name    	=> $name,
                __prg 		=> $prg,
                __param 	=> $param,
                __type 		=> $type,
                __input 	=> "",
                __output	=> "",
		__cmd  		=> "",
		__version 	=> $version,
		__virtual 	=> 0,
		__use_repository 	=>0,
		__path_repository 	=>"",
		__overwrite_repository	=>0
                };

    bless($self, $class);
# 	print "PROGRAM\tNew: $name, $prg, $param, $type, $version\n";
	$$R_LOGGER->Trace("PROGRAM\tNew: $name, $prg, $param, $type, $version\n",$LOGGER_STAMP,"D") if ( defined($R_LOGGER));
    return ($self);

}


=head2 procedure SetInOut

 Title        : SetInOut
 Usage        : $prog->SetInOut($input_file, $output_file);
 Prerequisite : $input_file exist
 Function     : Modify the attribute input and output of the Program object
 Returns      : none
 Args         : $input_file : the input file
 				$output_file : the output file
 Globals      : none

=cut 

sub SetInOut
{
    my ($self, $input_file, $output_file) = @_;

    $self->{'__input'} = $input_file;
    $self->{'__output'} = $output_file;
    $self->__BuildCmd;
}

=head2 procedure __BuildCmd

 Title        : __BuildCmd
 Usage        : __BuildCmd ;
 Prerequisite : All attribut (whithout $cmd) have to be set
 Function     : Private fonction which set the __cmd attribute
 Returns      : none
 Args         : none
 Globals      : none

=cut 

sub __BuildCmd
{
	my ($self) = @_;
	# Programmers should add some special concat data for a program depending on the name attribute.
	$self->{'__cmd'} = "$self->{'__prg'} $self->{'__param'} ";

	if ($self->{'name'} eq "blast")
	{	
		$self->{'__cmd'} .= "-i $self->{'__input'} -o $self->{'__output'}\n";
	}
	elsif ($self->{'name'} =~ /rnafold|rnaeval|b2ct/)
	{
		$self->{'__cmd'} .= "< $self->{'__input'} > $self->{'__output'}\n";
	}
	elsif ($self->{'name'} =~ /rnaalifold|ct2b|cmsearch/)
	{
		$self->{'__cmd'} .= "$self->{'__input'} > $self->{'__output'}\n";
	}
	elsif ($self->{'name'} eq "rnaplot")
	{
		$self->{'__cmd'} .= "< $self->{'__input'} \n";
	}
	elsif ($self->{'name'} eq "multalin")
	{
		$self->{'__cmd'} .= "$self->{'__input'} \n";
	}
	elsif ($self->{'name'} eq "mfold")
	{	
		my ($dir,$in) = $self->{'__input'} =~ /(.+)\/(.+)/;
		($dir,$in) = ('.',$self->{'__input'}) if ( $in eq '' );
		$self->{'__cmd'} = "(cd $dir; ".$self->{'__cmd'}." SEQ=$in 1>/dev/null ) 2>/dev/null\n";
	}
	elsif ($self->{'name'} eq "cmbuild")
	{	
		$self->{'__cmd'} .= "$self->{'__output'} $self->{'__input'} 1>/dev/null\n";
	}
	elsif ($self->{'name'} eq "formatdb")
	{	
		$self->{'__cmd'} = "$self->{'__prg'} $self->{'__input'} $self->{'__param'}\n";
	}
	elsif ($self->{'name'} eq "clustal")
	{	
		$self->{'__cmd'} .= "-INFILE=$self->{'__input'} -OUTFILE=$self->{'__output'} 1>/dev/null\n";
	}
	elsif ( $self->{'name'} =~ /rfam_scan|tRNAscan/i )
	{
		$self->{'__cmd'} .= "-o $self->{'__output'} $self->{'__input'} 2> /dev/null\n";
	}
	else
	{
		$self->{'__cmd'} .= "--output $self->{'__output'} --input $self->{'__input'} \n";
	}

}

=head2 procedure Run

 Title        : Run
 Usage        : $prog->Run ;
 Prerequisite : $cmd have to be set
 Function     : fonction which launch the command on system and use repository if __load_from_repository is =1
 Args         : $format_out : gff or else if you want to convert result in a specific format

=cut 
sub Run
{
    my ($self, $format_out) = @_;
    #print "PRG ".$self->GetName()." ".$self->{'__cmd'}."\n";

    # Si resultat dans le repository :
    if ($self->{'__use_repository'})
    {   
	my ($md5upper) = &GetMD5($self->{'__input'});
	# Verifie que le resultat est dans le repository
	my ($precomputed) = &GetFromRepositoryIfAlreadyComputed($md5upper, $self->{'__output'} ,$self->{'name'}."_".$self->{'__version'}, $self->{'__repository_dir'});		
	$$R_LOGGER->Trace("\tProgram ".$self->{'name'}."\tRun\tGetFromRepositoryIfAlreadyComputed: ".$precomputed." \n",$LOGGER_STAMP,"D") if ( defined($R_LOGGER));
		
        if (!$precomputed || $self->{'__overwrite_repository'})
        {	
	        # Si le resultat n'est pas dans le repository, on calcul et on sauvegarde
		$$R_LOGGER->Trace("\tProgram\tRun\tComputeAndSaveAnalysis: \n",$LOGGER_STAMP,"D") if ( defined($R_LOGGER));
        	&ComputeAndSaveAnalysis($md5upper, $self->{'__output'}, $self->{'__input'}."[._]*", $self->{'__cmd'},$self->{'name'}."_".$self->{'__version'},$self->{'__repository_dir'}, $self->{'__overwrite_repository'});
        }
    }
    else 
    {	
		$$R_LOGGER->Trace("\tProgram\tRun>".$self->{'name'}."_".$self->{'__version'}."<\tCMD: ".$self->{'__cmd'}." \n",$LOGGER_STAMP,"D") if ( defined($R_LOGGER));
		system ("$self->{'__cmd'} 2> /dev/null");
		# AUTRE erreur ici si rfam_scan intern
		#print STDERR $self->{'__cmd'}."\n";
		#print q "RUN2\n";
    }

    if (defined($format_out) && $format_out eq 'gff')
	{
		$self->ConvertToGff();
	}
	$$R_LOGGER->Trace("\tProgram\tRun>".$self->{'__version'}."<\t".$self->{'__cmd'}."\n",$LOGGER_STAMP,"D") if ( defined($R_LOGGER));

}


=head2 procedure SetLoadFromRepository

 Title        : SetLoadFromRepository
 Usage        : $prog->SetLoadFromRepository() ;
 Prerequisite : none
 Function     : Set the flag __load_from_repository at 1 to use the repository

=cut 

sub SetRepository
{ 
	my ($self, $dir,$ow_dir)=@_;
	$self->{'__use_repository'}=1 ;
	$self->{'__repository_dir'}=$dir ;
	$self->{'__overwrite_repository'}=$ow_dir ;
}

=head2 procedure UnsetRepository

 Title        : UnsetLoadFromRepository
 Usage        : $prog->UnsetLoadFromRepository() ;
 Prerequisite : none
 Function     : Set the flag __load_from_repository at No to not use the repository
 Returns      : none
 Args         : none
 Globals      : none

=cut 

sub UnsetLoadFromRepository
{ 
	my ($self)=@_;
	$self->{'__use_repository'}=0 ;
}



=head2 function ParseResult

 Title        : ParseResult
 Usage        : $prog->ParseResult ;
 Prerequisite : $cmd have to been run, parse only result of spécifique programs
 Function     : return interesting data from result file, depending on the program
 Returns      : data

=cut 
sub ParseResult
{my ($self) = @_;
	if ($self->{'name'} =~ /rnafold|rnaploteval|ct2b/)
	{
		return $self->__ParseFold();
	}
	elsif ($self->{'name'} eq "rnaalifold")
	{
		return $self->__ParseAliFold();
	}
	elsif ($self->{'name'} eq "cmsearch")
	{	
		return $self->__ParseCmsearch();
	}
}

=head2 function GetCmd

 Title        : GetCmd
 Usage        : $prog->GetCmd ;
 Function     : To get the string command
 Returns      : a string

=cut 
sub GetCmd
{
    my ($self) = @_;
    return $self->{'__cmd'};
}

=head2 function SetCmd

 Title        : SetCmd
 Usage        : $prog->SetCmd($cmd) ;

=cut 
sub SetCmd
{
    my ($self,$cmd) = @_;
    $self->{'__cmd'}=$cmd;
}

=head2 function ConvertToGff

 Title        : ConvertToGff
 Usage        : $prog->ConvertToGff ;
 Prerequisite : Program had been run
 Function     : Convert output file of Program to gff file

=cut 
sub ConvertToGff
{
    my ($self) = @_;
	# coverti dans le format : # <seqname> <source> <feature> <start> <end> <score> <strand> <frame> Att="Attributes"
	if (-e $self->{'__output'})
	{
		if ($self->{'name'} eq "trnascan")
		{
			$self->__ConvertTrnaInGff;
		}
		elsif ($self->{'name'} eq "blast")
		{
			$self->__ConvertBlastInGff;
		}
		elsif ($self->{'name'} =~ /rfam_scan/)
		{
			$self->__ConvertRfamScanInGff;		
		}
		#$self->SetFileOut($self->GetFileOut().".gff");
	}
}


=head2 function __ConvertRfamScanInGff

 Title        : __ConvertRfamScanInGff
 Usage        : __ConvertRfamScanInGff()
 Prerequisite : __output is set and file exists
 Function     : Transform output of rfam_scan in gff file


=cut 
sub __ConvertRfamScanInGff
{
	my ($self) = @_;

	my $rfamscan_file=$self->{'__output'};
	my $output_gff_file=$self->{'__output'}.$Var::E_GFF;
	$self->SetFileOut( $output_gff_file);
	my ($line, @a_tmp, $i, $gff_line );
	open (F_RFAM, $rfamscan_file) or die "Program->__ConvertRfamScanInGff : Impossible d'ouvrir: $rfamscan_file\n";
	open (F_OUT_GFF,">$output_gff_file")  or die "Program->__ConvertTrnaInGff : Impossible d'ouvrir/ecrire: $output_gff_file\n";
	# <seqname> <source> <feature> <start> <end> <score> <strand> <frame> [attributes] [comments]
	while ($line=<F_RFAM>)  
	{	
		chomp($line);
		next if ($line eq "" || $line =~ /^[\/s]$/);
		@a_tmp=split(/[\s]+/, $line);
		next if ($#a_tmp <10  || $line =~ /^#/ );
		$a_tmp[0] =~ s/ /_/;
		$gff_line = $a_tmp[0]."\t".$a_tmp[1]."\t".$a_tmp[2]."\t".$a_tmp[3]."\t".$a_tmp[4]."\t".$a_tmp[5]."\t".$a_tmp[6]."\t".$a_tmp[7]."\tAtt ".'"'.$a_tmp[8].$a_tmp[9].$a_tmp[10].'"';
		print F_OUT_GFF "$gff_line\n";

	}
	close F_RFAM;
	close F_OUT_GFF;
}


=head2 function __ConvertTrnaInGff

 Title        : __ConvertTrnaInGff
 Usage        : __ConvertTrnaInGff()
 Prerequisite : __output is set and file exists
 Function     : Transform output of tRNAscan-SE in gff file

=cut 

sub __ConvertTrnaInGff
{
	my ($self) = @_;

	my $trna_file=$self->{'__output'};
	my $output_gff_file=$self->{'__output'}.$Var::E_GFF;
	$self->SetFileOut( $output_gff_file);
	return if (not -e $trna_file);
	my ($line, @a_tmp, $i, $gff_line );
	open (F_TRNA, $trna_file) or die "Program->__ConvertTrnaInGff : Impossible d'ouvrir: $trna_file\n";
	open (F_OUT_GFF,">$output_gff_file")  or die "Program->__ConvertTrnaInGff : Impossible d'ouvrir/ecrire: $output_gff_file\n";
	# lecture des trois 1ere lignes
	for $i (0 .. 2)
		{$line = <F_TRNA>;}
	# <seqname> <source> <feature> <start> <end> <score> <strand> <frame> [attributes] [comments]

	while ($line=<F_TRNA>)  
	{	
		chomp($line);
		@a_tmp=split(/[\s]+/, $line);
		next if ($#a_tmp ==-1);
		$a_tmp[0] =~ s/ /_/;
		my ($beg,$end,$brin) = ( $a_tmp[2] < $a_tmp[3] ) ? ( $a_tmp[2],$a_tmp[3],'+' ) : ( $a_tmp[3],$a_tmp[2],'-' );
		$gff_line = $a_tmp[0]."\ttrnascan\ttRNA\t$beg\t$end\t.\t$brin\t.\tAtt ".'"'.$a_tmp[4].",".$a_tmp[5].",SCORE=".$a_tmp[8].'"';
		print F_OUT_GFF "$gff_line\n";

	}
	close F_TRNA;
	close F_OUT_GFF;
}


=head2 function __ConvertBlastInGff

 Title        : __ConvertBlastInGff
 Usage        : __ConvertBlastInGff()
 Prerequisite : __output is set and file exists
 Function     : Transform output of blast in gff file

=cut 

sub __ConvertBlastInGff
{
	my ($self) = @_;

	my $blast_file=$self->{'__output'};
	my $output_gff_file=$self->{'__output'}.$Var::E_GFF;
	$self->SetFileOut( $output_gff_file);
	my ($line, @a_tmp, @a_ac_id, $i, $gff_line );
	open ( F_BLAST, $blast_file) or die "ConvertBlastInGff : impossible d'ouvrir >$blast_file<" ;
	open ( F_OUT_GFF,">$output_gff_file") or die "ConvertBlastInGff : impossible d'ouvrir >$output_gff_file<" ;
	# Fields: Query id, Subject id, % identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score
	while ($line = <F_BLAST>)  
	{	
		chomp($line);
		next if ($line eq "");
		@a_tmp=split(/[\s]+/, $line);
		next if ($#a_tmp ==-1);
		$a_tmp[0] =~ s/ /_/;
		@a_ac_id=split(/;/, $a_tmp[1]); #ex : AC151964.12/37619-37798;RFL00001;RFL00001;
		if (not defined($a_ac_id[1]))
		{
			$a_ac_id[1]=$a_ac_id[0];
			$a_ac_id[2]="";
		}
		else
		{
			$a_ac_id[1].=";";
		}
		my ($beg,$end,$brin);
		if ( $a_tmp[6]<$a_tmp[7]) # si query est sur + 
		{	# verifie que subject est +/-
			($beg,$end,$brin) = ( $a_tmp[8] < $a_tmp[9] ) ? ( $a_tmp[6],$a_tmp[7],'+' ) : ( $a_tmp[6],$a_tmp[7],'-' );
		}
		else
		{
			($beg,$end,$brin)=( $a_tmp[7],$a_tmp[6],'-' )
		}
		$gff_line = $a_tmp[0]."\tblast\tHSP\t$beg\t$end\t.\t$brin\t.\tAtt ".'"'."ACC=".$a_ac_id[0].",ID=".$a_ac_id[1].$a_ac_id[2].",SCORE=".$a_tmp[11].'"';
		print F_OUT_GFF "$gff_line\n";
		#print "$gff_line\n";		
	}
	close F_BLAST;
	close F_OUT_GFF;
}


=head2 function ConvertToClustal

 Title        : ConvertToClustal
 Usage        : ConvertToClustal($clustal_file)
 Prerequisite : __output is set and file exists and is in multalin format
 Function     : Transform output of mutalin in clustal format

=cut 
sub ConvertToClustal
{	print "ConvertToClustal\n";
	my ($self, $clustal_file) = @_;
	my $ma_file=$self->{'__output'};
	# on ajoute un traitement perl car reformat, laisse des espace devant les accession => pas correct pour certain prg.
	my $cmde="$Var::CMD_REFORMAT $Var::REFORMAT_OUT_CLU".$ma_file;
	$cmde=$cmde ." | perl -ne \'{split;if ( \$#_==1 ) { \$_=sprintf(\"%-20s%s\\n\",\$_[0],\$_[1]); }; print; }\' > ".$clustal_file;
	system ($cmde);
}



=head2 procedure __ParseFold

 Title        : __ParseFold
 Usage        : __ParseFold();
 Prerequisite : Rnafold was exec and output file exist
 Function     : Parse result of rnafold
 Returns      : structure, free_energy


=cut 

sub __ParseFold
{
	my ($self) = @_;
	#ex :
	#>Acc
	#AGUGUCACCUUUGUGCUGUGCACUGUUGCAUGUCGCUUUUGCAUGGGCUGGUGCACCCCAC
	#....)).))))).))).(((((((..(.(((((.......))))).)..)))))))..... (-55.00)

	open(FIN, $self->GetFileOut()) or die "ParseFold : impossible d'ouvrir >". $self->GetFileOut()."<\n" ;	
	my $line;
	my $free=0;
	my $struct="";
	
	while ($line=<FIN>)
	{
		#next if ($line =~ /^[\>_ACGTUN]/) ;
		next if ($line !~ /^[\.\(\)]/) ;
		chomp($line);
		my @a_F=split(" ", $line, 2);
		my ($tmp_free) = $a_F[1]=~/\((.+)\)/;# recupère l'energie libre
		$free+=$tmp_free;
		$struct.=$a_F[0];
		
	}

	close FIN;
	return ($struct,$free);

}
=head2 procedure __ParseAliFold

 Title        : __ParseAliFold
 Usage        : __ParseAliFold();
 Prerequisite : Rnaalifold was exec and output file exist
 Function     : Parse result of rnaAlifold
 Returns      : structure, free_energy, longueur, consensus

=cut 

sub __ParseAliFold
{
	my ($self) = @_;
	# ex : peux ^etre sur plusieur lignes !!!!
	#GUGUGCCUGGCUCCCUGUAUGCCAUUUGUAGAGUCCAUCACGACAGUGAUGAACUUAGUGAAUGGCGUACUCGGAGUCAUGCAUGC
	#((((((.(((((((..((((((((((..(.((((.((((((....)))))).)))).)..))))))))))..))))))).))))))
	#minimum free energy = -50.20 kcal/mol (-50.20 +   0.00)
	
	my ($fig_name, $len, $cons, $line);
	open(FIN, $self->GetFileOut())  or die "ParseAliFold : impossible d'ouvrir >". $self->GetFileOut()."<\n" ;	
	  $cons=<FIN>; # ligne séquence
	  chomp($cons);
	  $len=length($cons);
	  $line=<FIN>; # ligne de structure secondaire
	  chomp($line);				
	  my @a_F=split(/ /, $line, 2);
	  my ($free) = $a_F[1]=~/\((.+)=/;# recupère l'energie libre
	close FIN;
	# retourn : structure, free_energy, longueur, consensus
	return ($a_F[0],$free,$len, $cons)
}

=head2 procedure __ParseCmsearch

 Title        : __ParseCmsearch
 Usage        : __ParseCmsearch();
 Prerequisite : Cmsearch was exec and output file exist
 Function     : Parse result of cmsearch
 Returns      : 1 if have a match, 0 else.

=cut 

sub __ParseCmsearch
{
	my ($self) = @_;
	#ex :
	#sequence: AC128638_r1
	#hit 0   :      1     86    15.46 bits
	#           <<<<<<-<<<<<<<-<<<<<<<<<--<<<_______________________________
	#         1 uugugcCuggcuccCuguaugccaCacguguAGacCaAcaaauguuuuggaaauguCuaC 60
	#           :UGUGCCUG:CUCCC G:AUGCCA +::U +AG CCA CA      U    +AUG    +
	#         1 GUGUGCCUGGCUCCCUGUAUGCCAUUUGUAGAGUCCAUCACGACAGU----GAUGAACUU 56


	my ($res) = 0;
	open (FIN, $self->GetFileOut()) or die "ParseCmsearch : impossible d'ouvrir >". $self->GetFileOut()."<\n" ;;	
	while (my $line = <FIN>)
	{
		if ($line =~ /^sequence:/)
		{
			# traite cherche hit !!!0
			$line = <FIN>;
			if ($line =~ /^hit/)
			{
				$res = 1;
				
			}
		}
	}
	close (FIN);
	return ($res);
}


=head2 function GetFileOut

 Title        : GetFileOut
 Usage        : GetFileOut()
 Function     : Accessor
 Returns      : output file path

=cut 
sub GetFileOut
{
	my ($self) = @_;

	return $self->{'__output'};
}

=head2 function SetFileOut

 Title        : SetFileOut
 Usage        : SetFileOut()
 Function     : accessor
 Returns      : modify output file path

=cut 

sub SetFileOut
{
	my ($self, $output) = @_;

	$self->{'__output'}=$output;
}

=head2 function GetName

 Title        : GetName
 Usage        : GetName()
 Function     : Name accessor
 Returns      : name of program
 
=cut 
sub GetName
{
	my ($self) = @_;

	return $self->{'name'};
}

=head2 function GetVersion

 Title        : GetVersion
 Usage        : GetVersion()
 Function     : Accessor
 Returns      : version

=cut 
sub GetVersion
{
	my ($self) = @_;

	return $self->{'__version'};
}

=head2 function SetVersion

 Title        : SetVersion
 Usage        : SetVersion()
 Function     : accessor
 Returns      : modify version

=cut 
sub SetVersion
{
	my ($self, $version) = @_;

	$self->{'__version'}=$version;
}

=head2 function GetType

 Title        : GetType
 Usage        : GetType()
 Function     : Accessor
 Returns      : type

=cut 
sub GetType
{
	my ($self) = @_;

	return $self->{'__type'};
}

=head2 function SetPrg

 Title        : SetPrg
 Usage        : SetPrg()
 Function     : accessor
 Returns      : modify version

=cut 
sub SetPrg
{
	my ($self, $prg) = @_;

	$self->{'__prg'}=$prg;
}


=head2 function GetPrg

 Title        : GetPrg
 Usage        : GetPrg()
 Function     : accessor

=cut 
sub GetPrg
{
	my ($self) = @_;

	return ($self->{'__prg'});
}


=head2 function SetParam

 Title        : SetParam
 Usage        : SetParam()
 Function     : accessor
 Returns      : modify version

=cut 
sub SetParam
{
	my ($self, $param) = @_;

	$self->{'__param'}=$param;
}


=head2 function GetParam

 Title        : GetParam
 Usage        : GetParam()
 Function     : accessor

=cut 
sub GetParam
{
	my ($self) = @_;

	return ($self->{'__param'});
}

=head2 function SetVirtual

 Title        : SetVirtual
 Usage        : SetVirtual($flag)
 Returns      : modify version

=cut 
sub SetVirtual
{
	my ($self, $flag) = @_;

	$self->{'__virtual'}=$flag;
}


=head2 function IsVirtual

 Title        : IsVirtual
 Usage        : IsVirtual()
 Function     : accessor

=cut 
sub IsVirtual
{
	my ($self) = @_;

	return ($self->{'__virtual'});
}

####################################################
#################  function MD5  ###################
####################################################
=item GetFromRepositoryIfAlreadyComputed

=cut 

sub GetFromRepositoryIfAlreadyComputed
{
	my ($md5upper, $output_file , $analysis, $repository_dir ) = @_;
		
	my $ext_ok=0;
	my ($dirin) = "$repository_dir/$analysis";
	if (!-d $dirin) {return 0};
	my ($output_dir, $file_name) = $output_file =~ /(.+)\/(.+)/; 
	my ($file) = "$dirin/$md5upper.$file_name";
		my $cde;
	if (-e $file)
	{
		$cde = "cp -a $file $output_file";
		$ext_ok= 1;
	}
    	elsif (-e "$file$Var::E_GZ")
    	{
    		$cde = "cp -a $file$Var::E_GZ $output_file$Var::E_GZ ; gzip -fd $output_file$Var::E_GZ";
	    	$ext_ok= 1;
	}

	if ($ext_ok)
	{
		
  		$cde .= " ; cd $output_dir ; tar xf $output_file ; cd $repository_dir; " if ( $file =~ /\.tar$/ );

		system ($cde);	
   		#print "Analysis=$analysis: $output_file downloaded from repository ($dirin)\n";
    		return 1;
	}
    	return 0;
}

=item ComputeAndSaveAnalysis

=cut 
	
sub ComputeAndSaveAnalysis
{
    my ($md5upper,$output_file, $input_file, $cmd , $analysis, $repository_dir, $overwrite_repository) = @_;
    my ($cde) = "";
    if ($overwrite_repository)
    {
	system "rm -f $output_file";
    }

    my ($res) = system $cmd;
    &Save($md5upper, $output_file, $input_file, $analysis, $repository_dir);
 
}

=item GetMD5

=cut

sub GetMD5
{
    my ($seqfasta) = @_;
    my ($md5upper) = ('');

    my @a_seq = `grep -v \\> $seqfasta`;
    chop(@a_seq);
    my $seq = join('', @a_seq);
    $seq =~ s/\W//g;
    $md5upper = md5_hex("\U$seq");
    return ($md5upper);
}


=item Clean

=cut

sub Clean
{
    my ($id, $seqdata, $analysis) = @_;
    my ($file) = "$seqdata.$analysis";
    system ("rm -f $file");
    #print("$analysis($id) removed");
}

=item Save

=cut 

sub Save
{
    	my ($md5upper, $output, $input, $analysis, $analysis_repository) = @_;
    	my ($dirout) = "$analysis_repository/$analysis";
	if (!-d $dirout)
	{    
		system "mkdir -p $dirout";	
	}
	
	my ($dir, $file_name) = $output =~ /(.+)\/(.+)/;
	$md5upper = "$md5upper." if ($md5upper ne "" && $md5upper !~ /\.$/);
	my ($cde);
	
	if( -e  $output )
	{
		$cde="cp -a $output $dirout/$md5upper$file_name ; gzip -f $dirout/$md5upper$file_name";
	}
	else 
	{
	    	my ($input_dir, $input_file) = $input =~ /(.+)\/(.+)/;
		$cde="cd $input_dir ; tar -czf  $dirout/$md5upper$file_name.gz $input_file; cd $dirout";
	}
    	system ($cde);

    #print ("Analysis=$analysis($output) computed and saved in the repository ($dirout)\n");
}

1;

