#
# $Id: Runner.pm 35 2005-11-21 16:09:58Z manu $
#

package Runner;

# Copyright Notice

# Copyright I.N.R.A. - C.N.R.S. 

# emmanuel.courcelle@toulouse.inra.fr
# jerome.gouzy@toulouse.inra.fr

# This software is a perl module whose purpose is to help you writing 
# your own scripts

# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# http://www.cecill.info" 

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

Runner - Run an external program through fork, with timeout and other signal management

=head1 NOT AN OBJECT

This package is NOT an object oriented package

=head1 EXPORTED FUNCTIONS

=head2 RunExt

 my ($status, $killed) = RunExt(-cmd $cmd,
                                -stdin  $stdin,
				 -stdout $stdout,
				 -stderr $sterr,
				 -timeout $timeout);

 PARAMETERS:

   -cmd     = The external command whith its parameters.
   -stdin   = The standard input, if redirected   (a file name, or anything you can pass to open)
   -stdout  = The standard output, if redirected  (a file name, or anything you can pass to open)
   -stderr  = The standard error, if redirected   (a file name, or anything you can pass to open)
   -timeout = The timeout in seconds. No timeout if -1


 RETURNED VALUES:
   $status  ==> the return value of the child
   $timeout ==> 0 if OK
                -1 if the child was killed by timeout
		-2 if the child was killed by TERM (ctrl-C, qdel, ...)

The cmd is executed, it may be killed if a TERM signal or an ALMR is received. 
The child is then killed with a TERM, followed by a KILL signal: it should return in any circumstance

=head2 RunExtNoWait

 my $pid = RunExtNoWait(-cmd $cmd,
                        -stdin  $stdin,
			 -stdout $stdout,
			 -stderr $sterr,
			 -timeout $timeout);

 PARAMETERS:

   -cmd     = The external command whith its parameters.
   -stdin   = The standard input, if redirected   (a file name, or anything you can pass to open)
   -stdout  = The standard output, if redirected  (a file name, or anything you can pass to open)
   -stderr  = The standard error, if redirected   (a file name, or anything you can pass to open)


 RETURNED VALUE:
   $pid ==> the process Id of the child

The cmd is executed, but the sub returns without waiting for the child's death.
The children started may be killed if a TERM signal or an ALARM signal are received.
The child is then killed with a TERM signal. NO KILL signal is sent to the child,
terminating the child in a graceful way. If I don't have to wait, I am less violent...

=cut

use strict;
use warnings;
use Carp;
use IO::File;

BEGIN
{
    use Exporter ();
    use vars qw($VERSION @ISA @EXPORT);

    our $VERSION = do {my @r = (q$Rev: 207 $ =~ /\d+/g); $r[0]};
    @ISA         = qw(Exporter);
    @EXPORT      = qw(&RunExt &RunExtNoWait);
    $SIG{'TERM'} = sub {};          # Ignoring the SIGINT
    $SIG{'ALRM'} = \&KillProcess;
}

our $RUN_PID   = 0;     # Only one process ran by RunExt
our @RUN_PID   = ();    # Many processes ran by RunExtNoWait
our $KILLED_BY = '';

#
# EXPORTED function
#

sub RunExt
{
    my %args = @_;
    croak("Runner.pm internal error (RUN_PID $RUN_PID)") unless ($RUN_PID == 0);
    croak("The cmd to run is NOT specified")             unless (exists($args{'-cmd'}));

    if ($KILLED_BY eq 'TERM')
    {
        return (0, -2);    # a TERM signal was received JUST BEFORE calling RunExt
    }
    else
    {
        $KILLED_BY = '';
    }

    my @arguments = split(' ', $args{'-cmd'});

    if (exists $args{'-timeout'})
    {                      # Program a timeout, if necessary
        my $timeout = $args{'-timeout'} + 0;
        if ($timeout > 0)
        {
            alarm($args{'-timeout'});
        }
    }
    $RUN_PID = fork();

    if ($RUN_PID == 0)
    {                      # T H E   C H I L D
        my ($path, @arguments) = split(' ', $args{'-cmd'});
        if (exists $args{'-stdin'})
        {
            close(STDIN);
            my $stdin = $args{'-stdin'};
            open(STDIN, $stdin) or die "CANNOT OPEN $stdin";
        }
        if (exists $args{'-stdout'})
        {
            close(STDOUT);
            my $stdout = $args{'-stdout'};
            open(STDOUT, $stdout) or die "CANNOT OPEN $stdout";
        }
        if (exists $args{'-stderr'})
        {
            close(STDERR);
            my $stderr = $args{'-stderr'};
            open(STDERR, $stderr) or die "CANNOT OPEN $stderr";
        }
        exec($path, @arguments);    # should NOT return from here
        exit 1;                     # If returns, something bad happened
    }

    my $rvl = waitpid($RUN_PID,0);    # T H E   P A R E N T
    
    my $sts = $? >> 8;
    $RUN_PID = 0;

    if ($KILLED_BY eq 'ALRM')
    {          # The child was killed by a timeout
        return (0, -1);
    }
    elsif ($KILLED_BY eq 'TERM')
    {          # The child was killed by a TERM signal (qdel)
        return (0, -2);
    }
    else
    {          # The child returned, with or without an error
        return ($sts, 0);
    }
}

#
# EXPORTED function
#

sub RunExtNoWait
{
    my %args = @_;
    croak("The cmd to run is NOT specified") unless (exists($args{'-cmd'}));

    if ($KILLED_BY eq 'TERM')
    {
        return 0;    # a TERM signal was received JUST BEFORE calling RunExtNoWait
    }
    else
    {
        $KILLED_BY = '';
    }

    my @arguments = split(' ', $args{'-cmd'});

    my $run_pid = fork();

    if ($run_pid == 0)
    {                # T H E   C H I L D
        my ($path, @arguments) = split(' ', $args{'-cmd'});
        if (exists $args{'-stdin'})
        {
            close(STDIN);
            my $stdin = $args{'-stdin'};
            open(STDIN, $stdin) or die "CANNOT OPEN $stdin";
        }
        if (exists $args{'-stdout'})
        {
            close(STDOUT);
            my $stdout = $args{'-stdout'};
            open(STDOUT, $stdout) or die "CANNOT OPEN $stdout";
        }
        if (exists $args{'-stderr'})
        {
            close(STDERR);
            my $stderr = $args{'-stderr'};
            open(STDERR, $stderr) or die "CANNOT OPEN $stderr";
        }
        exec($path, @arguments);
        exit 1;
    }

    push(@RUN_PID, $run_pid);

    return $run_pid;

}

#
# Title: KillProcess
# Usage: kill the process whose PID is in $RUN_PID.
#        Send him a TERM signal, then a KILL signal
#        kill the processes whose PID are in @RUN_PID, and remove them from @RUN_PID
#        Send them ONLY a TERM signal (less violence...).
#        Called by the ALRM and the TERM signal handlers
#

sub KillProcess
{
    my $signame = shift;
    if ($RUN_PID != 0)
    {
        kill 'TERM', $RUN_PID;
        kill 'KILL', $RUN_PID;

        #	warn "$$ killed $RUN_PID";
    }
    while (my $pid = shift(@RUN_PID))
    {
        kill 'TERM', $pid;

        #  	warn "$$ killed $pid";
    }
    $KILLED_BY = $signame;
}

1;

