#
# $Id$
#

# Copyright INRA

# Celine.Noirot@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr
# Christine.Gaspin@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based system for detecting, clustering and annotating non protein
# coding RNA genes.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
 
#
# 	Celine.Noirot@toulouse.inra.fr
#	Last updated: October 21, 2005
#

package Sequence;

=pod

=head1 NAME

Sequence- It's an object Sequence

=head1 SYNOPSIS

	my $seq = New Sequence($seq_file);	
	
example: 
	use Sequence;
	my $seq= New Sequence();
	$seq->Initialize("/cmain/LeARN/data/pipeline/data.fa");
	$seq->Print();
	print $seq->GetSubSequence(50,59,"+")."\n";
	print $seq->GetSubSequence(50,59,"-")."\n";
=head1 DESCRIPTION


=cut

use strict;
use warnings;
use Bio::SeqIO;
use File::stat;
use Time::localtime;
#      $date_string = ctime(stat($file)->mtime);
#       print "file $file updated at $date_string\n"

BEGIN
{
    our $VERSION = "0.2";
}

our $STEP_RNA_ID=1; # default 1 but can be change by class method SetStepRnaId

=head2 function New

 Title        : New
 Usage        : New Sequence;
 Prerequisite : none
 Function     : Create an object Sequence
 Returns      : An object Sequence
 Args         : none
 Globals      : none

=cut 

sub New
{
    my ($class, $seq_file) = @_;

    my $self = {
		__id   		=> "",
                __acc 		=> "",
                __version 	=> "",
		__length 	=> "0",
                __organism	=> "",
                __last_upd	=> "",
		__seq_file	=> "",
		__sequence	=> "",
		__ra_rna	=> [],
		__next_rna_id=> 1,
		__analyzed	=> 0,
		__old_bac_acc => ""
                };

    bless($self, $class);

	if (defined $seq_file)
	{
		$seq_file=~s/^\s//;
		$seq_file=~s/\s$//;
		$self->SetSeqFile($seq_file);
#		$self->SetSequence();
		$self->SetLastUpd(ctime(stat($seq_file)->mtime)) if (-e $seq_file);
	}
	return ($self);

}


=head2 procedure Initialize

 Title        : Initialize
 Usage        : $seq->Initialize($id,$acc, $version, $organism, $last_upd, $seq_file);
 Prerequisite : none
 Function     : Initialize the Sequence object
 Returns      : none
 Args         : none
 Globals      : none

=cut 

sub Initialize
{
	my ($self, $id,$acc, $version, $organism, $last_upd, $seq_file) = @_;
	$self->SetId($id);
	$self->SetAcc($acc);
	$self->SetVersion($version);
	$self->SetOrganism($organism);
	if (defined $last_upd){$self->SetLastUpd($last_upd);}	
	if (defined $seq_file){$self->SetSeqFile($seq_file);}
	#$self->SetSequence(); # get the sequence in an attribute
}


=head2 fuction GetSubSequence

 Title        : GetSubSequence
 Usage        : $seq->GetSubSequence($start, $end);
 Prerequisite : none
 Function     : Return a piece of sequence
 Returns      : none
 Args         : none
 Globals      : none

=cut 

sub GetSubSequence
{
	my ($self, $start, $end, $strand) = @_;
	my $sub_seq;
	if (not defined($self->{'__sequence'}) || $self->{'__sequence'} eq "")
	{
		$sub_seq=undef;
	}
	else
	{
		$sub_seq= substr($self->GetSequence(), $start-1, abs ($end-$start)+1);
		if ($strand eq '-')
		{
			$sub_seq = reverse($sub_seq);
    	    		$sub_seq =~ tr/ATGCatgc/TACGTACG/;
		}
		$sub_seq =~ tr/Tt/Uu/;
	}
	return uc($sub_seq);
}

=head2 fuction DumpAnnotation

 Title        : DumpAnnotation
 Usage        : $seq->DumpAnnotation();
 Prerequisite : none
 Function     : Save in a file all Rna of this sequence.
 Args         : none

=cut 

sub DumpAnnotation
{
	my ($self, $file) = @_;

	if (not defined($file))
	{
		$file = $self->{'__seq_file'}.".learn.gff";
	}
	my $line="";
	open (GFF,">$file");
	foreach my $i (0 .. $#{$self->{'__ra_rna'}})
	{
		$line=${$self->{'__ra_rna'}[$i]}->GetGffStringFromSeq($self->GetId());
		print GFF $line."\n";
		print STDERR ">$line<\n";
	}
	close (GFF);
	
}

=head2 fuction Print

 Title        : Print
 Usage        : $seq->Print;
 Prerequisite : none
 Function     : Print
 Returns      : none
 Args         : none
 Globals      : none

=cut 

sub Print
{
	my ($self) = @_;
	print "\t--->Sequence\n";
	print "\tId : ".$self->GetId()."\n";
	print "\tAcc : ".$self->GetAcc()."\n";
	print "\tVersion : ".$self->GetVersion()."\n";	
	print "\tOrganism : ".$self->GetOrganism()."\n";
	print "\tLast upd : ".$self->GetLastUpd()."\n";
	print "\tSeq file : ".$self->GetSeqFile()."\n";
	print "\tNext rna id : ".$self->GetNextRnaId()."\n";
}
######### ACCESSEUR ########"

=item GetId

	Return the Id of a sequence (accession_number.version)
=cut 

sub GetId
{
	my ($self) = @_;

	return $self->{'__id'};
}

=item SetId

	Set the Id of a sequence (accession_number.version)
=cut 

sub SetId
{
	my ($self, $id) = @_;

	$self->{'__id'}=$id;
}

=item GetAcc

	Return the acc of a sequence 
=cut 

sub GetAcc
{
	my ($self) = @_;
	return $self->{'__acc'};
}

=item SetAcc

	Set the acc of a sequence 
=cut 

sub SetAcc
{
	my ($self, $acc) = @_;
	$self->{'__acc'}=$acc;
}

=item GetVersion

	Return the version of a sequence 
=cut

sub GetVersion
{
	my ($self) = @_;
	return $self->{'__version'};
}

=item SetVersion

	Set the version of a sequence 
=cut

sub SetVersion
{
	my ($self, $version) = @_;
	$self->{'__version'}=$version;
}


=item GetOrganism

	Return the organism of a sequence 
=cut

sub GetOrganism
{
	my ($self) = @_;

	return $self->{'__organism'};
}

=item SetOrganism

	Set the organism of a sequence 
=cut

sub SetOrganism
{
	my ($self, $organism) = @_;

	$self->{'__organism'}=$organism;
}

=item GetLastUpd

	Return the date of last update
=cut

sub GetLastUpd
{
	my ($self) = @_;

	return $self->{'__last_upd'};
}

=item SetLastUpd

	Set the date of last update
=cut

sub SetLastUpd
{
	my ($self, $last_udp) = @_;

	$self->{'__last_upd'}=$last_udp;
}

=item GetSeqFile

	Return the path of the sequence file
=cut

sub GetSeqFile
{
	my ($self) = @_;

	return $self->{'__seq_file'};
}

=item SetSeqFile

	Set the path of the sequence file
=cut

sub SetSeqFile
{
	my ($self, $seq_file) = @_;

	$self->{'__seq_file'}=$seq_file;
}

=item SetPathOfSeqFile

	Set the path of the sequence file (whithout changing the name of file.
=cut

sub SetPathOfSeqFile
{
	my ($self, $path) = @_;

	$self->{'__seq_file'}=$path."/".$self->{'__id'}."/".$self->{'__id'};
}

=item GetLength

	Get the length of the sequence data
=cut

sub GetLength
{
	my ($self) = @_;
	return $self->{'__length'};
}

=item GetSequence

	Get the sequence (data) 
=cut


sub GetSequence
{
	my ($self) = @_;

	return $self->{'__sequence'};
}

=item UnsetSequence

	Free the sequence (data) field (take too much place in memory) 
=cut

sub UnsetSequence
{
	my ($self) = @_;
	$self->{'__sequence'}=undef;
}

=item SetSequence

	Open sequence file and initialize sequence data field
=cut

sub SetSequence
{
	my ($self) = @_;
	if (not -e $self->{'__seq_file'})
	{
#		$$Learn::R_LOGGER->Trace("\tSequence\tSetSequence\tFile not found ".$self->{'__seq_file'}." \n",$Learn::LOGGER_STAMP,"DE");
		die ("Sequence file not found : ##".$self->{'__seq_file'}."##\n");
	}
	else
	{
		my $in = Bio::SeqIO -> new( -file => $self->{'__seq_file'}, '-format' => 'Fasta' ) ;
		my $seq = $in->next_seq();
		$self->{'__sequence'}=uc($seq->seq());
		$self->{'__length'}=length($self->{'__sequence'});
		$in->DESTROY();
	}
}

=item GetNextRnaId

	Return the next number for the annotation of a rna on the sequence 
=cut

sub GetNextRnaId
{
	my ($self) = @_;

	return $self->{'__next_rna_id'};
}

=item SetNextRnaId

	Set the next number for the annotation of a rna on the sequence 
=cut

sub SetNextRnaId
{
	my ($self, $next_rna_id) = @_;

	$self->{'__next_rna_id'}=$next_rna_id;
}

=item SetStepRnaId

	Class method.
	Initialize the step of rna id on the sequence (change for all sequences)
=cut

sub SetStepRnaId
{
	my ($class, $value) = @_;
	$STEP_RNA_ID=$value;
}

=item RegisterRna

	When a rna is validated, init the next rna id from the rna accession
=cut

sub RegisterRna
{
	my ($self,$rna_id) = @_;

	my ($num)= $rna_id=~/([0-9]*)$/;
	$self->SetNextRnaId($num+$STEP_RNA_ID);
}

=item GetAnalyzed

	Get the flag to know if the sequence have been analysed in an old release
=cut

sub GetAnalyzed
{
	my ($self) = @_;

	return $self->{'__analyzed'};
}

=item GetAnalyzed

	Set Flag to know if the sequence have been analysed in an old release
=cut

sub SetAnalyzed
{
	my ($self, $analyzed) = @_;

	$self->{'__analyzed'}=$analyzed;
}



=item GetIsNewVersion

	Get the flag to know if the sequence is a new version of an existing sequence in LeARN
=cut

sub GetOldBacAcc
{
	my ($self) = @_;

	return $self->{'__old_bac_acc'};
}

=item SetIsNewVersion

	Set Flag to know if  the sequence is a new version of an existing sequence in LeARN
=cut

sub SetOldBacAcc
{
	my ($self, $old_bac) = @_;

	$self->{'__old_bac_acc'}=$old_bac;
}

=item GetRna

	Return a reference on an rna at the index wich is on this sequence.
=cut

sub GetRna
{
	my ($self, $index) = @_;
	return $self->{'__ra_rna'}[$index];
}

=item AddRna

	Add an  rna's reference on the array of rna
=cut

sub AddRna
{
	my ($self, $r_rna) = @_;
	push (@{$self->{'__ra_rna'}}, $r_rna);

}

=item DeleteRna

	Delete an rna of the array of rna at the index $index
	$sequence->DeleteRna($index)
	
=cut

sub DeleteRna
{
	my ($self, $index) = @_;
	splice (@{$self->{'__ra_rna'}}, $index,1);
}

=item NumberOfRna

	Return the number of rna
	
=cut

sub NumberOfRna
{
	my ($self) = @_;
	return $#{$self->{'__ra_rna'}}+1;
}
1;

