#
# $Id$
#

# Copyright INRA

# Celine.Noirot@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr
# Christine.Gaspin@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based system for detecting, clustering and annotating non protein
# coding RNA genes.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
 
#
# 	Celine.Noirot@toulouse.inra.fr
#	Last updated: October 21, 2005
#

package Rna;

=pod

=head1 NAME

Rna-It's an object Rna
	Rna create its own SecondaryStructure, and several Range object. 
	Need a reference of a program, for computing SecondaryStructure.

=head1 SYNOPSIS

	my $prog = New RNA();	
	
example: 
	use Rna;
	my $my_rna = New Rna();
	$my_rna->Initialize();
	$my_rna->SetPrincipalRange();
	$my_rna->ComputeStructure();

=head1 DESCRIPTION


=cut

use strict;
use warnings;
use SecondaryStructure;
use Range;
use Learn;
use UtilRnaml;
use Annotate;
BEGIN
{
    our $VERSION = "0.2";


}

use Var;
#use vars qw($Var::DIR_IMG $Var::DIR_TMP  $Var::DIR_SS $Var::E_CT $Var::E_IMG $Var::E_XML );

my %H_RNAML;

=head2 function New

 Title        : New
 Usage        : New Rna;
 Prerequisite : none
 Function     : Create an object Rna
 Returns      : An object
 Args         : none
 Globals      : none

=cut 

sub New
{
	my ($class) = @_;
	
	my $self = {
		id    			=> "",
		__name 			=> "",
		__desc  		=> "",
		__last_upd		=> " ",
		__stucture		=> New SecondaryStructure(),
		__principal_range 	=> ref,
		__a_ranges 		=> [],
		__status		=> "",
		__path			=> "",
		__dbxref		=> "",
		__r_prog_for_structure 	=>'@',
		__family_acc 		=>"",
		__a_annotate 		=> [],
                };

    	bless($self, $class);
	$$Learn::R_LOGGER->Trace("Rna\tNew\n",$Learn::LOGGER_STAMP,"D");
    	return ($self);
}	


=head2 procedure Initialize

 Title        : Initialize
 Usage        : $rna->Initialize();
 Prerequisite : none
 Function     : Initialize the Rna object
 Returns      : none
 Args         : $id
 				$name
				$desc
				$last_upd
				$status : 
				$path : directory of image
				$r_prog_struct: reference on a program to compute the structure
 Globals      : none

=cut 

sub Initialize
{
    my ($self,$id, $name, $desc, $last_upd, $status, $path, $r_prog_struct, $r_prog_convert, $family) = @_;
	$$Learn::R_LOGGER->Trace("Rna\tInitialize\tParameters:$id\n",$Learn::LOGGER_STAMP,"D");
	$self->SetId($id);
	$self->SetName($name);
	$self->SetDesc($desc);
	$self->SetLastUpd($last_upd);
	$self->SetStatus($status);
	$self->SetPath($path);
	$self->SetProgForStructure($r_prog_struct);
	$self->{'__stucture'}->SetProgram($r_prog_struct);
	$self->{'__stucture'}->SetProgramConvert($r_prog_convert);
	$self->{'__stucture'}->SetImage("$path/$Var::DIR_IMG/$id$Var::E_IMG");
	$self->{'__stucture'}->SetSSFile("$path/$Var::DIR_SS/$id$Var::E_CT");
	$self->{'__stucture'}->SetRnamlFile("$path/$Var::DIR_SS/$id$Var::E_XML");
	#$self->ComputeNameDbxref($desc);
	$self->SetFamilyAcc($family) if defined ($family);
	
}




=head2 procedure ComputeNameDbxref

 Title        : ComputeNameDbxref
 Usage        : $rna->ComputeNameDbxref($desc);
 Prerequisite : none
 Function     : Split desc and extract data depending on program which detect rna
 Returns      : none
 Args         : $desc :  description
 Globals      : none

=cut 

sub ComputeNameDbxref
{
	# decoupe le champ description 
	# Rfam : 	ACC=RF00001;ID=5S_rRNA;SCORE=127.09;
	# Blast: ACC=RF60000;ID=msx1-as;n1200.1/0-2184;SCORE=1732
	# tRna : Ala;GGC;SCORE=44.80
	# iscan : ACC=RFL00001;ID=RFL00001;SCORE=127.09;
	my ($self,$desc, $feature)=@_;
	my $dbxref="-";
	my $name="-";
	my @a_T=split(/[;,=]/, $desc); 
	if ($#a_T > -1)
	{	
		if ( $a_T[0] =~ /^ACC/)
		{
			if ($a_T[1] =~ /^RF0/ )
			{
				$dbxref="Rfam_".$a_T[1];
			}
			elsif ($a_T[1] =~ /^RF5/ )
			{
				$dbxref="AddonRfam_".$a_T[1];
			}
			elsif ($a_T[1] =~ /^RF6/ )
			{
				$dbxref="BigRnaDb_".$a_T[1];
			}
			elsif ($a_T[1] =~ /^RF7/ )
			{
				$dbxref="BigRnaDb_".$a_T[1];
			}	
			elsif ($a_T[1] =~ /^RFL/ )
			{
				$dbxref="LeARN_".$a_T[1];
				$self->SetFamilyAcc($a_T[1]);
			}						
			$name=$a_T[3];
		}

		if ( defined ($feature) && $feature =~/trna/i )
		{
			$dbxref="tRNA_".$a_T[0];
			$name=$dbxref;
		}
                elsif ( defined ($feature) &&  $feature =~/snorna/i )
		{
			$dbxref="snorna";
			$name=$dbxref;
		}
	}
	$self->SetDbxref($dbxref);
	$self->SetName($name);

}

=head2 procedure Print

 Title        : Print
 Usage        : $rna->Print();
 Prerequisite : none
 Function     : Display object data
 Returns      : none
 Args         : none
 Globals      : none

=cut 

sub Print
{
	my ($self, $log)=@_;
	print "<br>RNA :\n";
	print "id : ".$self->GetId()."\n";
	print "Name : ".$self->GetName()."\n";
	print "Desc : ".$self->GetDesc()."\n";
	print "LastUpd : ".$self->GetLastUpd()."\n";

	print "FamilyAcc : ".$self->GetFamilyAcc()."\n";
	print "Structure : \n";
	$self->{'__stucture'}->Print();
	print "Prog strucut : ".${$self->{'__r_prog_for_structure'}}->GetName()."\n" if  (defined($self->{'__r_prog_for_structure'}));
	print "Path : ".$self->GetPath()."\n";
	print "-->RangePrincipal : \n";
	$self->{'__principal_range'}->Print();
	print "-->RangeSecondaire : $#{$self->{'__a_ranges'}}\n";
	for (my $i=0; $i<=$#{$self->{'__a_ranges'}}; $i++)
	{
		$self->{'__a_ranges'}[$i]->Print();
	}
	print "-->Annotation : \n";
	for (my $i=0; $i<=$#{$self->{'__a_annotate'}}; $i++)
	{
		$self->{'__a_annotate'}[$i]->Print();
	}
	if (defined($log))
	{
		$$Learn::R_LOGGER->Trace("Rna\tPrint\tid : ".$self->GetId()."\n",$Learn::LOGGER_STAMP,"D");
		$$Learn::R_LOGGER->Trace("Rna\tPrint\tName : ".$self->GetName()."\n",$Learn::LOGGER_STAMP,"D");
		$$Learn::R_LOGGER->Trace("Rna\tPrint\tDesc : ".$self->GetDesc()."\n",$Learn::LOGGER_STAMP,"D");
		$$Learn::R_LOGGER->Trace("Rna\tPrint\tLastUpd : ".$self->GetLastUpd()."\n",$Learn::LOGGER_STAMP,"D");
		$$Learn::R_LOGGER->Trace("Rna\tPrint\tStructure : \n",$Learn::LOGGER_STAMP,"D");
		$self->{'__stucture'}->Print("log");
		$$Learn::R_LOGGER->Trace("Rna\tPrint\tPath : ".$self->GetPath()."\n",$Learn::LOGGER_STAMP,"D");
		$self->{'__principal_range'}->Print("log");
		for (my $i=0; $i<=$#{$self->{'__a_ranges'}}; $i++)
		{
			$self->{'__a_ranges'}[$i]->Print("log");
		}

	}
}

=head2 procedure IsSame

 Title        : IsSame
 Usage        : $prog->IsSame($r_rna);
 Prerequisite : none
 Function     : Compare self with another Rna.
 Returns      : 		

				0 if not same
				1 if same sequence data
				2 if redondancy on same sequence (overlap) => on jete !
 Args         : $r_rna : reference on a  rna to compare
 Globals      : none

=cut 

sub IsSame
{
	my ($self,$r_rna)=@_;
	my $i=0;
	my $find=0;
	# verifier sequence mere iden fait ds le range
	# Compare le range principal
	# range retourne 2 si chevauchement / 1 si sequence data identique / 0 rien en commun
	my $res = $self->GetPrincipalRange()->IsSame($$r_rna->GetPrincipalRange());
	if ( $res == 0 )  # si pas de correspondance sur range primaire on regarde avec les ranges secondaires
	{
		while ($i<=$#{$self->{'__a_ranges'}} && $res==0)
		{
			$res=$self->{'__a_ranges'}[$i]->IsSame($$r_rna->GetPrincipalRange());
			$i++;
		}
	}
	# on retourne le type de chevauchement détecté
	return ($res);
}


=head2 procedure SaveRnaml

 Title        : SaveRnaml
 Usage        : $prog->SaveRnaml($dir);
 Prerequisite : none
 Function     : Write in directory a Rnaml file
 Args         : $dir : directry where rnaml file will be create
 Globals      : none

=cut 

sub SaveRnaml
{
	my ($self,$xml_dir,$tmp_dir, $rh_sequences, $rh_programs)=@_;
	$$Learn::R_LOGGER->Trace("Rna\tSaveRnaml\t".$self->GetId()."\tDir: $xml_dir\n",$Learn::LOGGER_STAMP,"DE");
	UtilRnaml::SaveRna($self,$xml_dir,$tmp_dir, $rh_sequences, $rh_programs);
}


=head2 procedure LoadFromRnaml

 Title        : LoadFromRnaml
 Usage        : $prog->LoadFromRnaml($file);
 Prerequisite : none
 Function     : Read a Rnaml file and initialize rna fields
 Args         : $file : rnaml file
 Globals      : none

=cut 

sub LoadFromRnaml
{
	my ($self,$xml_file,$dir_rna_release, $rh_sequences )=(@_);
	$$Learn::R_LOGGER->Trace("Rna\tLoadFromRnaml\t$xml_file\n",$Learn::LOGGER_STAMP,"DE");
	# rempli l'objet self avec les donnée dans le fichier rnaml
	&UtilRnaml::LoadRna(\$self,$xml_file,$dir_rna_release, $rh_sequences);

	# déplace les fichiers de l'ancienne release vers la nouvelle et modifie les attribut !
	my $old_path_img = $self->{'__stucture'}->GetImage();
	if (defined ($old_path_img) && ! -d $old_path_img && -e $old_path_img)
	{
		#ex : /www/LeARN/data/rel_0.1/db/learn/rna/img/AC148607_r1.jpg
		my @a_path=split('/',$old_path_img);
		$self->{'__stucture'}->SetImage("$dir_rna_release/$Var::DIR_IMG/".$a_path[$#a_path]);
		if ($old_path_img ne $self->{'__stucture'}->GetImage())
		{
			my $cmd="cp -a $old_path_img ".$self->{'__stucture'}->GetImage();
			system $cmd;
			$$Learn::R_LOGGER->Trace("Rna\tLoadFromRnaml\t$cmd\n",$Learn::LOGGER_STAMP,"DE");
		}
	}

	my $old_path_ss = $self->{'__stucture'}->GetSSFile();
	if (defined ($old_path_ss) && $old_path_ss ne "" && ! -d $old_path_ss && -e $old_path_ss)
	{
		#ex : /www/LeARN/data/rel_0.1/db/learn/rna/ss/AC148607_r1.rnaml
		my @a_path=split('/',$old_path_ss); # le dernier element de a_path contiens le nom du fichier
		$self->{'__stucture'}->SetSSFile("$dir_rna_release/$Var::DIR_SS/".$a_path[$#a_path]);
		if ($old_path_ss ne $self->{'__stucture'}->GetSSFile())
		{
			my $cmd="cp -a $old_path_ss ".$self->{'__stucture'}->GetSSFile();
			$$Learn::R_LOGGER->Trace("Rna\tLoadFromRnaml\t$cmd\n",$Learn::LOGGER_STAMP,"DE");
			system $cmd;
		}
	}
	
	my $old_path_rnaml = $self->{'__stucture'}->GetRnamlFile();
	if (defined ($old_path_rnaml)  && $old_path_rnaml ne ""  && ! -d $old_path_rnaml && -e $old_path_rnaml)
	{
		#ex : /www/LeARN/data/rel_0.1/db/learn/rna/ss/AC148607_r1.rnaml
		my @a_path=split('/',$old_path_rnaml); # le dernier element de a_path contiens le nom du fichier
		$self->{'__stucture'}->SetRnamlFile("$dir_rna_release/$Var::DIR_SS/".$a_path[$#a_path]);
		if ($old_path_rnaml ne $self->{'__stucture'}->GetRnamlFile())
		{
		my $cmd="cp -a $old_path_rnaml ".$self->{'__stucture'}->GetRnamlFile();
		$$Learn::R_LOGGER->Trace("Rna\tLoadFromRnaml\t$cmd\n",$Learn::LOGGER_STAMP,"DE");
		system $cmd;
		}
	}
}

=head2 procedure DumpFasta

 Title        : DumpFasta
 Usage        : $rna->DumpFasta();
 Prerequisite : none
 Function     : Write fasta file 
 Args         : optionnaly : $file path and file name you want to create !
 Globals      : none

=cut 

sub DumpFasta
{
	my ($self,$file)=@_;

	if (not defined($file))
	{
		system "mkdir -p ".$self->GetPath()."/$Var::DIR_TMP/".$self->GetId();
		$file = $self->GetPath()."/$Var::DIR_TMP/".$self->GetId()."/".$self->GetId()."$Var::E_FA";
	} 
	else
	{
		system "rm -rf ".$self->GetPath()."/$Var::DIR_TMP/".$self->GetId()."/*";
	}
	$$Learn::R_LOGGER->Trace("Rna\tDumpFasta\t".$self->GetId()."\tFile : $file\n",$Learn::LOGGER_STAMP,"DE");
	my $seq = Bio::Seq->new(-id => $self->GetId(), -seq => $self->{'__principal_range'}->GetSeqData());
	my $out = Bio::SeqIO->new( -file => ">$file", '-format' => 'Fasta' );
	$out->write_seq($seq);
	$out->DESTROY();
	return ($file);

}

=head2 procedure DumpSeqData

 Title        : DumpSeqData
 Usage        : $rna->DumpSeqData();
 Prerequisite : none
 Function     : Write sequence (on one line) in a file
 Args         : optionnaly : $file path and file name you want to create !
 Globals      : none

=cut 

sub DumpSeqData
{
	my ($self,$file,$struct)=@_;

	my $outdir = $self->GetPath()."/$Var::DIR_TMP/".$self->GetId();
	if ($file eq '' || ! -e $file)
	{
		system "mkdir -p $outdir";
		$file = "$outdir/".$self->GetId().".seq";
	}
	else
	{
		system "rm -rf $outdir/*";
	}
	$$Learn::R_LOGGER->Trace("Rna\tDumpSeqData\t".$self->GetId()."\tFile : $file\n",$Learn::LOGGER_STAMP,"DE");
	open OUT, ">$file" or die "LeARN : unable to create >$file<\n";
	print OUT ">".$self->GetId()."\n";
	print OUT $self->{'__principal_range'}->GetSeqData()."\n";
	if (defined ($struct) && $struct ne "")
	{
		print OUT $struct."\n";
	}
	close OUT;
	return ($file);
}


=head2 procedure GetGffStringFromSeq

 Title        : GetGffStringFromSeq
 Usage        : $rna->GetGffStringFromSeq($ref_seq);
 Prerequisite : none
 Function     : return a gff line coresponding to the range of the sequence in parameter 
 Args         : optionnaly : $ref_seq 
 Globals      : none

=cut 

sub GetGffStringFromSeq
{
	my ($self,$seq)=@_;
	my $line="";
	my $comment="";
	if (${$self->{'__principal_range'}->GetSequence()}->GetId() eq $seq)
	{
		$line=$self->{'__principal_range'}->GetGffFeature($self->GetId());
		$comment="FAM_ACC=".$self->GetFamilyAcc().";RNA_ACC=".$self->GetName();
	}
	else
	{
		foreach my $i (0..$#{$self->{'__a_ranges'}})
		{
			if (${$self->{'__a_ranges'}[$i]->GetSequence()}->GetId() eq $seq)
			{
				$line=$self->{'__principal_range'}->GetGffFeature($self->GetId());
				$comment="FAM_ACC=".$self->GetFamilyAcc().";RNA_ACC=".$self->GetName();
				last;
			}
		}
	}
	return ("$line\tAtt \"$comment\"");
}

=head2 procedure ClearTmpDir

 Title        : ClearTmpDir
 Usage        : $rna->ClearTmpDir();
 Prerequisite : none
 Function     : Delete tempory file of rna

=cut 
sub ClearTmpDir
{
	my ($self)=@_;
	my $tmp_dir = $self->GetPath()."/$Var::DIR_TMP/".$self->GetId();
	$$Learn::R_LOGGER->Trace("Rna\tClearTmpDir\t $tmp_dir\n",$Learn::LOGGER_STAMP,"DE");
	system "rm -rf $tmp_dir/*";
}

#########ACCESSEUR ########"



##### length #####
=head2 function GetLength

 Title        : GetLength
 Usage        : $rna->GetLength();
 Prerequisite : Length was set
 Returns      : Length of rna (principal range)
=cut 
sub GetLength
{
	my ($self) = @_;
	return $self->{'__principal_range'}->GetLength();
}


##### __stucture #####
=head2 function GetStructure

 Title        : GetStructure
 Usage        : $rna->GetStructure();
 Returns      : Objet structure
=cut 

sub GetStructure
{
	my ($self) = @_;
	return $self->{'__stucture'};
}

=head2 procedure SetStructure

 Title        : SetStructure
 Usage        : $rna->SetStructure($struct);
 Function     : Set the objet structure
 Args         : $struc an object structure
=cut 

sub SetStructure
{
	my ($self, $struct) = @_;

	$self->{'__stucture'}=$struct;
}


=head2 function ComputeStructure

 Title        : ComputeStructure
 Usage        : $rna->ComputeStructure();
 Function     : Launch funtion Compute on structure object , wich exec the fold program
 	
=cut 
sub ComputeStructure
{
	my ($self) = @_;

	my $file;
	my $type=${$self->{'__stucture'}->GetProgram()}->GetType();
	if ( ${$self->{'__stucture'}->GetProgram()}->IsVirtual() )
	{
		$self->{'__stucture'}->SetProgram(&Learn::GetRefProgram( $type, "actual"))
	}
	if ($type =~ /R_PLOT|R_PLOT_EVAL|R_FOLD_ALT/)
	{
		my $struct="";
		$struct=$self->GetStructure()->GetStructure() if ($type =~ /R_PLOT|R_PLOT_EVAL/);
		$file=$self->DumpSeqData("",$struct);
	}
	else 
	{
		$file=$self->DumpFasta();
	}
	
	my $res=($self->{'__stucture'}->Compute($file,$self->GetPath(), $self->GetId()));
	#my $cmd = "rm -f $file \n";
	#$$Learn::R_LOGGER->Trace("Rna\tComputeStructure\tCMD:".$cmd."\n",$Learn::LOGGER_STAMP,"D");
	#system $cmd;
	return $res;
}

=head2 procedure ComputePlotEval

 Title        : ComputePlotEval
 Prerequis    : Length of consensus and the structure have to be the same
 Usage        : $family->ComputePlotEval();
 Prerequis    : the parenthsis structure of structure object and consensus must be set
 Function     : Plot and eval free energy according to the parenthesis structure
 
=cut

sub ComputePlotEval
{
	my ($self,$consensus)=(@_);
	my $tmp_dir=$self->GetPath()."/tmp/".$self->GetId();
	mkdir $tmp_dir;
	my $tmp_file="$tmp_dir/".$self->GetId().".tmp";
	$consensus =~ s/[-_]/N/g;
	open ( SEQ , ">$tmp_file") or die "unable to create >$tmp_file<\n";
	print SEQ ">".$self->GetId()."\n";
	print SEQ $consensus."\n";
	print SEQ $self->GetStructure()->GetStructure()."\n";
	close (SEQ);
	
	my $save_prog=$self->GetStructure()->GetProgram();
	my $r_prog_plot_eval=&Learn::GetRefProgram('R_PLOT_EVAL', "actual");
	$self->{'__stucture'}->SetProgram($r_prog_plot_eval);
	my ($len, $cons) = $self->{'__stucture'}->Compute($tmp_file, $self->GetPath(), $self->GetId() );
	$self->{'__stucture'}->SetProgram($save_prog);

	print STDOUT "<br>Maintenant $consensus ($len, $cons) $tmp_dir / $tmp_file\n";
	
}


##### principal_range #####
=head2 function GetPrincipalRange

 Title        : GetPrincipalRange
 Usage        : $rna->GetPrincipalRange();
 Returns      : The object principal range
 	
=cut 
sub GetPrincipalRange
{
	my ($self) = @_;

	return $self->{'__principal_range'};
}

=head2 procedure SetPrincipalRange

 Title        : SetPrincipalRange
 Usage        : $rna->SetPrincipalRange($start, $end, $strand, $r_sequence, $r_prog_detect,$date_detect, $seq_data, $length);
 Function     : Create a new range. Si l arn a deja un range principal ce dernier passe en range secondaire
 Args         :	$start 	
				$end
				$strand
				$r_sequence : reference on a sequence object
				$r_prog_detect : reference on the program wich detect this range
				$date_detect : date of execution of prog_detect
				[$seq_data] : if set $seq_data : no computtion from sequence
				[$length] : if set $length idem .
=cut 
sub SetPrincipalRange
{
	my ($self, $start, $end, $strand, $r_sequence, $r_prog_detect,$date_detect, $seq_data, $length) = (@_);
	$$Learn::R_LOGGER->Trace("Rna\tSetPrincipalRange\t".$self->GetId().": $start, $end, $strand\n",$Learn::LOGGER_STAMP,"D");
	if (ref($self->{'__principal_range'}))
	{
		$self->AddSecondaryRange( 
			$self->{'__principal_range'}->GetStart(),
			$self->{'__principal_range'}->GetEnd(),
			$self->{'__principal_range'}->GetStrand(),
			$self->{'__principal_range'}->GetSequence(), 
			$self->{'__principal_range'}->GetProgDetect(),
			$self->{'__principal_range'}->GetDateDetect(),
			$self->{'__principal_range'}->GetSeqData(),
			$self->{'__principal_range'}->GetLength()		); #sauvegarde l'ancien range principal
			
	}
	$self->{'__principal_range'}=New Range();
	if (defined($seq_data))
	{
	$self->{'__principal_range'}->InitAll($start, $end, $strand, $r_sequence, $r_prog_detect,$date_detect,$seq_data, $length);
	}
	else
	{
	$self->{'__principal_range'}->Initialize($start, $end, $strand, $r_sequence, $r_prog_detect,$date_detect);
	}
}


=item GetId

	Get the id of the rna
	
=cut 

sub GetId
{
	my ($self) = @_;

	return $self->{'id'};
}

=item SetId

	Set the id of the rna
	
=cut 

sub SetId
{
	my ($self, $id) = @_;

	$self->{'id'}=$id;
}

=item GetName

	Get the name of the rna
	
=cut 

sub GetName
{
	my ($self) = @_;

	return $self->{'__name'};
}

=item SetName

	Set the name of the rna
	
=cut 

sub SetName
{
	my ($self, $name) = @_;
	$self->{'__name'}=$name;
}

=item GetDesc

	Get the description of the rna
	
=cut 

sub GetDesc
{
	my ($self) = @_;

	return $self->{'__desc'};
}

=item SetDesc

	Set the description of the rna
	
=cut 

sub SetDesc
{
	my ($self, $desc) = @_;

	$self->{'__desc'}=$desc;
}

=item GetLastUpd

	Get the last update of the rna
	
=cut 

sub GetLastUpd
{
	my ($self) = @_;

	return $self->{'__last_udp'};
}

=item SetLastUpd

	Set the last update of the rna
	
=cut 

sub SetLastUpd
{
	my ($self, $last_udp) = @_;

	$self->{'__last_udp'}=$last_udp;
}

=item GetSequence

	Get the seq data of the principal range of the rna
	
=cut 

sub GetSequence
{
	my ($self) = @_;
	return $self->{'__principal_range'}->GetSeqData();
}

=item GetStatus

	Get the status of the rna
	
=cut 

sub GetStatus
{
	my ($self) = @_;

	return $self->{'__status'};
}

=item SetStatus

	Set the status of the rna
	
=cut 

sub SetStatus
{
	my ($self, $status) = @_;

	$self->{'__status'}=$status;
}

=item GetSecondaryRange

	Get an object secondary range at the index 
	$rna->GetSecondaryRange($index);
=cut 

sub GetSecondaryRange
{
	my ($self, $index) = @_;
	return $self->{'__a_ranges'}[$index];
}

=item GetNumberOfSecondaryRange

	Get the number of  secondary range
	
=cut 

sub GetNumberOfSecondaryRange
{
	my ($self) = @_;
	return $#{$self->{'__a_ranges'}};
}

=item AddSecondaryRange

	Add a secondary range
	$rna->AddSecondaryRange($start, $end, $strand, $r_sequence, $r_prog_detect,$date_detect, [ $seq_data, $length] )
=cut 

sub AddSecondaryRange
{
	my ($self, $start, $end, $strand, $r_sequence, $r_prog_detect,$date_detect, $seq_data, $length) = @_;
	$$Learn::R_LOGGER->Trace("Rna\tAddSecondaryRange\t".$self->GetId().": $start, $end, $strand\n",$Learn::LOGGER_STAMP,"D");
	my $range = New Range();
	if (defined($seq_data))
	{
		$range->InitAll($start, $end, $strand, $r_sequence, $r_prog_detect, $date_detect, $seq_data, $length);
	}
	else
	{
	$range->Initialize($start, $end, $strand, $r_sequence,$r_prog_detect,$date_detect);
	}
	push (@{$self->{'__a_ranges'}},$range);
	
}

=item DeleteSecondaryRange

	Delete a secondary range at an index
	$rna->DeleteSecondaryRange($index);
=cut 


sub DeleteSecondaryRange
{
	my ($self, $index) = @_;
	splice (@{$self->{'__a_ranges'}}, $index,1);
}

=item GetPath

	Get the path o the rna
	
=cut 

sub GetPath
{
	my ($self) = @_;

	return $self->{'__path'};
}

=item SetPath

	Set the path o the rna
	
=cut 

sub SetPath
{
	my ($self, $path) = @_;

	$self->{'__path'}=$path;
}


=item GetProgForStructure

	Get the reference on object program uses for computing secondary structure
	
=cut 

sub GetProgForStructure
{
	my ($self) = @_;
	return $self->{'__r_prog_for_structure'};
}

=item SetProgForStructure

	Set the reference on object program uses for computing secondary structure
	And also set this program to the object secondary structure
	
=cut 

sub SetProgForStructure
{
	my ($self, $prog) = @_;
	$self->{'__r_prog_for_structure'}=$prog;
	$self->{'__stucture'}->SetProgram($prog);
}

=item GetFamilyAcc

	Get the family accession of the rna
	
=cut 

sub GetFamilyAcc
{
	my ($self) = @_;
	return $self->{'__family_acc'};
}

=item SetFamilyAcc

	Set the family accession of the rna
	
=cut 

sub SetFamilyAcc
{
	my ($self, $name) = @_;
	$self->{'__family_acc'}=$name;
}

=item GetDbxref

	Get the dbxref attribute
	
=cut 

sub GetDbxref
{
	my ($self) = @_;
	return $self->{'__dbxref'};
}

=item SetDbxref

	Set the dbxref attribute
	
=cut 

sub SetDbxref
{
	my ($self, $var) = @_;
	$self->{'__dbxref'}=$var;
}

=item GetAnnotate

	Get an annotation at the index $index
	$rna->GetAnnotate($index);
=cut 

sub GetAnnotate
{
	my ($self, $index) = @_;
	return $self->{'__a_annotate'}[$index];
}

=item GetIndexFromTimeStamp

	Get the index of an annotation according to a time stamp (use as an identifiant)
	$rna->GetIndexFromTimeStamp($time);
=cut 

sub GetIndexFromTimeStamp
{
	my ($self, $time) = @_;
	my $i=0;
	foreach my $annotate (@{$self->{'__a_annotate'}})
	{
		if ($annotate->GetTimeStamp() eq $time )
		{
			last;
		}
		$i++;
	}
	return $i;
}

=item AddAnnotate

	Add an object annotation 
	$rna->AddAnnotate($annotate);
	
=cut 

sub AddAnnotate
{
	my ($self, $annotate) = @_;
	push (@{$self->{'__a_annotate'}},$annotate);
}

=item DeleteAnnotate

	Delete an annotation at the index $index
	$rna->DeleteAnnotate($index);
	
=cut 

sub DeleteAnnotate
{
	my ($self, $index) = @_;
	
	splice(@{$self->{'__a_annotate'}},$index,1 );
}

=item GetNbAnnotations

	Get the number of annotation 
	
=cut 

sub GetNbAnnotations
{
	my ($self) = @_;
	return $#{$self->{'__a_annotate'}};
}

=item AddNewAnnotate

	Build a new annotation and add it to the array of annotation
	$rna->AddNewAnnotate( $start, $end,$title, $legend, $login, $time_stamp);
=cut 

sub AddNewAnnotate
{
	my ($self,  $start, $end,$title, $legend, $login, $time_stamp) = @_;
	push (@{$self->{'__a_annotate'}},New Annotate());
	my $annot=$self->GetAnnotate($self->GetNbAnnotations); # recup le dernier elment ajout
        $annot->Initialize($start, $end,$title, $legend, $login, $time_stamp);
}
1;

